import logging
import inspect

from abc import ABC
from typing import Any, Callable, Dict, List, Mapping, Optional, Set, Tuple, Type, TypeVar, Union
from typing import get_type_hints
from inds.core.exceptions import InvalidInputError, ImplementationError, MisconfiguredVariableError, ConfigError
from inds.core.config import get_implementation, get_implementations, is_user_defined_class, pretty_class_name

logger = logging.getLogger(__name__)

__NodeType = TypeVar('T', bound='Node')


def instantiate_from_config(config: Optional[dict], class_handle: Type[__NodeType], *args,
                            conditionals: Optional[Dict[str, Any]] = None, global_path: Optional[List] = None,
                            localns=None, return_only_params: Optional[Set[str]] = None, **kwargs) -> __NodeType:
    """

    :param config:
    :param class_handle:
    :param args:
    :param conditionals:
    :param localns:
    :param kwargs:
    :param global_path:        Path in the global configuration dictionary
    :param return_only_params: if not none, the names of the parameters that are to be parsed and returned.
    :return:
    """
    if ABC in class_handle.__bases__:
        raise ImplementationError(f'Trying to instantiate the abstract base class {class_handle.__name__}.')

    if global_path is None:
        global_path = []

    # Retrieve all parameters that are available to the class' constructor and erase those that are specified manually.
    all_parameters: List[str] = list(inspect.signature(class_handle.__init__, ).parameters)[1 + len(args):]
    for name in kwargs.keys():
        if name not in all_parameters:
            raise ImplementationError(f'Explicitly passed parameter {name} is not required. Abort.')
        if conditionals is not None and name in conditionals:
            raise ImplementationError(f'Passed parameter {name} as conditional and ordinary argument. Abort.')
        all_parameters.remove(name)

    # Keep content of conditionals that are accepted by the constructor, and remove from the list of all parameters.
    conditionals: Dict[str, Any] = {k: v for k, v in (conditionals or dict()).items() if k in all_parameters}
    for name in conditionals.keys():
        all_parameters.remove(name)

    if return_only_params is not None:
        all_parameters = [p for p in all_parameters if p in return_only_params]

    # Parse configuration options.
    parsed: Dict[str, Any] = dict()
    flat_config: Dict[str, Tuple[List[str], Any]] = dict()
    if config is not None:
        parsed, flat_config = parse_config_options(config, class_handle.__init__, list(all_parameters),
                                                   localns=localns, global_path=global_path)

    if return_only_params is not None:
        return parsed

    # Instantiate class from all parameters.
    try:
        return class_handle(*args, **parsed, **kwargs, **conditionals)
    except ConfigError as e:
        e.paths.extend([(flat_config.get(name)[0], e) for name in (e.names if e.names is not None else [])
                        if flat_config.get(name) is not None])
        e.names = None
        raise e from None


def parse_config_options(config: Dict[str, Any], function_ptr: Callable, variable_names: List[str],
                         localns=None, global_path: Optional[List] = None) -> Tuple[Dict[str, Any],
                                                                                    Dict[str, Tuple[List[str], Any]]]:
    global_path: List = global_path if global_path is not None else []

    default_params: Mapping[str, inspect.Parameter] = inspect.signature(function_ptr).parameters

    # Fix for being able to instantiate string objects in Cython.
    if hasattr(function_ptr, '__annotations__'):
        for k in function_ptr.__annotations__.keys():
            if function_ptr.__annotations__[k] == 'unicode':
                function_ptr.__annotations__[k] = 'str'

    # Remove variable arguments (**args, *kwargs) from the list of all parameters to be
    for var_arg in {name for name, val in default_params.items() if val.kind in (val.VAR_KEYWORD, val.VAR_POSITIONAL)}:
        logger.debug(f'Removing variable argument {var_arg} from default instantiation of {function_ptr}.')
        variable_names.remove(var_arg)

    type_hints: Dict[str, Type] = get_type_hints(function_ptr, localns=localns)  # Can throw a type error

    # In order to be able to flatten logical groups, expand configuration options that are unused in the constructor
    # and check them for missing configuration fields
    remaining_items: List[Tuple[List[str], Any]] = [([k], v) for k, v in config.items()]
    flat_config: Dict[str, Tuple[List[str], Any]] = dict()
    while len(remaining_items):
        key, value = remaining_items.pop(0)

        full_key: str = '_'.join(key)
        single_key: str = key[-1]

        found_key = [k for k in (full_key, single_key) if k in default_params]

        if len(found_key):
            if found_key[0] not in flat_config:
                flat_config[found_key[0]] = global_path + key, value
            else:
                logger.warning(f'Duplicate potential field {found_key[0]}')

        else:
            if type(value) == dict:
                remaining_items.extend((([str(f) for f in (*key, k) if f], v) for k, v in value.items()))

    # XXX: potentially summarize the configurations that fail in here.
    return_values: Dict[str, Any] = dict()
    failed_attributes: Dict[str, MisconfiguredVariableError] = dict()

    for variable_name in variable_names:

        try:
            # Fetch the default parameter if any
            default_parameter = default_params.get(variable_name)
            if default_parameter is not None and hasattr(default_parameter, 'default'):
                default_parameter = default_parameter.default

            if default_parameter == inspect.Parameter.empty:
                default_parameter = None

            # Fetch the user-defined value. In case the user did not specify the value in the config,
            # fall back to default.
            path, value = flat_config.get(variable_name, ([variable_name], default_parameter))

            # Check that the types match the required type.
            if variable_name in type_hints:
                type_hint = type_hints.get(variable_name)
                value = decode_value_type(value, type_hint, variable_name, localns=localns, global_path=global_path)

            return_values[variable_name] = value

        except MisconfiguredVariableError as e:
            failed_attributes[variable_name] = e

    if len(failed_attributes):
        raise MisconfiguredVariableError.combine(list(failed_attributes.values()))

    return return_values, flat_config


def decode_value_type(value: Any, type_hint: Type, name: str, global_path: List, localns=None):
    is_container: bool = hasattr(type_hint, '__args__')
    template_types: Tuple = getattr(type_hint, '__args__') if is_container else (type_hint,)
    if is_container and template_types is None:
        template_types = tuple()

    if type(value) == type_hint or type_hint == TypeVar or type_hint == Any or type(type_hint) == TypeVar:
        return value

    if type(value) is dict and type(value.get('selection')) is str and type_hint == str:
        return value.get('selection')

    try:

        # Instantiate list of something.
        if len(template_types) == 1 and type_hint == List[template_types]:
            if type(value) is not list:
                raise InvalidInputError(f'Expected type: list. Found type: {type(value)}.')
            return [decode_value_type(val, template_types[0], name, global_path=global_path + [i], localns=localns)
                    for i, val in enumerate(value)]

        # Instantiate dictionary key - value pairs.
        if len(template_types) == 2 and type_hint == Dict[template_types]:
            if type(value) is not dict and not isinstance(value, dict):
                raise InvalidInputError(f'Expected type: dict. Found type: {type(value)}.')

            return {
                decode_value_type(k, template_types[0], name, global_path=global_path + [k], localns=localns):
                    decode_value_type(v, template_types[1], name, global_path=global_path + [k], localns=localns)
                for k, v in value.items()
            }

        # Instantiate Tuple from tuple or list.
        if len(template_types) > 1 and type_hint == Tuple[template_types]:
            if not (type(value) is list or type(value) is tuple
                    and len(template_types) == len(value)):
                raise InvalidInputError(f'Expected type: list / tuple of length '
                                        f'{len(template_types)}. '
                                        f'Found type: {type(value)} with wrong length.')

            return tuple([decode_value_type(value[i], template_type, name, global_path=global_path + [i],
                                            localns=localns) for i, template_type in enumerate(template_types)])

        # Instantiate one of the provided types.
        if len(template_types) > 1 and type_hint == Union[template_types]:
            errors: List[MisconfiguredVariableError] = list()
            for template_type in template_types:
                try:
                    return decode_value_type(value, template_type, name, global_path=global_path, localns=localns)
                except MisconfiguredVariableError as e:
                    errors.append(e)
            raise MisconfiguredVariableError.combine(errors)

        #
        # Check if the value's type matches directly.
        #
        try:
            #if isinstance(type(value), type_hint):
            #    return value  # XXX Maybe this was here for a purpose.

            if isinstance(value, type_hint):
                return value

            # Convert int to float if necessary
            if type(value) == int and type_hint == float:
                return float(value)

            # Convert string to int or float if necessary and possible
            if type(value) == str and (type_hint == float or type_hint == int):
                return type_hint(value)

        except Exception:
            pass

        #
        # Try to instantiate the user-defined class.
        #

        if is_user_defined_class(type_hint):

            # Instantiate a user-defined class via the build method.
            build_error: Optional[Exception] = None
            has_build: bool = False

            if value is None:
                build_error = InvalidInputError('value is None')

            try:
                has_build = callable(getattr(type_hint, 'build', None))

                # If the user-defined type has a build method, try the build method can be used to
                # instantiate the instance of the class from the configuration
                if has_build and isinstance(value, dict):

                    # dictionary key-value pair, which creates the required name as key.
                    if 'name' in inspect.signature(type_hint.build).parameters and len(value) == 1 \
                            and type(list(value.keys())[0]) is str and type(list(value.values())[0]) is dict:
                        return type_hint.build(name=next(iter(value.keys())), config=next(iter(value.values())))

                    else:
                        return type_hint.build(config=value)
            except Exception as e:
                build_error = e

                try:
                    return type_hint.build(config=value)
                except Exception:
                    ...

            # Instantiate a user-defined class by calling the constructor directly.
            instantiation_error: Optional[Exception] = None
            if not has_build:
                try:
                    return type_hint(value)
                except Exception:
                    pass

                try:
                    return instantiate_from_config(value, type_hint, global_path=global_path, localns=localns)
                except Exception as e:
                    instantiation_error = e

            #
            # Try to instantiate a subclass
            #

            # Instantiation of subclass by name provided as string argument
            subclass_instantiation_error: Optional[Exception] = None
            try:
                if type(value) is str:
                    return get_implementation(type_hint, value)()

                # Instantiation of subclass with command line arguments.
                if type(value) is dict and len(value) == 1 and type(list(value.keys())[0] is str):
                    subclass = get_implementations(type_hint, list(value.keys())[0])
                    key_to_sub_value = list(value.keys())[0]
                    return decode_value_type(value[key_to_sub_value], subclass, name,
                                             global_path=global_path + [key_to_sub_value], localns=localns)

            except Exception as e:
                subclass_instantiation_error = e

            if type(value) is dict and type(value.get('selection')) is str:  # XXX:
                try:
                    subclass = get_implementation(type_hint, value.get('selection'))
                    return decode_value_type(value, subclass, name, global_path=global_path, localns=localns)

                except Exception as e:
                    subclass_instantiation_error = e
                    try:
                        return type_hint(value.get('selection'))
                    except Exception:
                        pass

            #
            # Raise the error that conveys most information
            #
            errors: Tuple[List[Exception], List[Exception]] \
                = (list(), list())
            if subclass_instantiation_error is not None:
                # Only set if there is a subclass which matches the string description.
                attribute = type(subclass_instantiation_error) is AttributeError
                if not isinstance(subclass_instantiation_error, MisconfiguredVariableError):
                    subclass_instantiation_error = InvalidInputError(
                        f'Subclass instantiation {str(subclass_instantiation_error)}')
                errors[attribute].append(subclass_instantiation_error, path=global_path)

            if build_error is not None:
                # Only set if a build function exists and the value appears to match the desired
                # format.
                attribute = type(build_error) is AttributeError
                if not isinstance(build_error, MisconfiguredVariableError):
                    build_error = InvalidInputError(f'Call to build(...) {str(build_error)}')
                errors[attribute].append(build_error, path=global_path)

            if instantiation_error is not None:
                attribute = type(instantiation_error) is AttributeError
                if not isinstance(instantiation_error, MisconfiguredVariableError):
                    instantiation_error = InvalidInputError(f'Direct instantiation: {str(instantiation_error)}')
                errors[attribute].append(instantiation_error, path=global_path)

            if len(errors[0]):
                raise errors[0][0]

            if len(errors[1]):
                raise errors[1][0]

            raise InvalidInputError('No suitable configuration option found.')

        raise InvalidInputError('No valid configuration option detected.')
    except MisconfiguredVariableError as e:
        e.append(name, pretty_class_name(type_hint), value, path=global_path)
        raise e

    except Exception as e:
        raise MisconfiguredVariableError(name, pretty_class_name(type_hint), value, str(e), path=global_path)
