import numpy as np

from inds.core.decorators import numba


@numba.jit(numba.boolean(numba.float32[::1], numba.float32[::1], numba.float32), nopython=True)
def iou(box1, box2, iou_threshold):
    """
    Filter based on the IoU of both provided boxes.
    :param box1: First box for comparison (x, y, width, height).
    :param box2: Second box for comparison (x, y, width, height).
    :param iou_threshold: Threshold of the intersection of union (IoU) of both boxes.
    :return: Boolean if boxes overlap more than @param iou_threshold.
    """

    x1 = box1[0]
    y1 = box1[1]
    w1 = box1[2]
    h1 = box1[3]

    x2 = box2[0]
    y2 = box2[1]
    w2 = box2[2]
    h2 = box2[3]

    dx = np.abs(x1 - x2)
    dy = np.abs(y1 - y2)

    w_intersect = np.minimum(np.maximum(0.0, 0.5 * w1 + 0.5 * w2 - dx), np.minimum(w1, w2))
    h_intersect = np.minimum(np.maximum(0.0, 0.5 * h1 + 0.5 * h2 - dy), np.minimum(h1, h2))

    area1 = np.multiply(w1, h1)
    area2 = np.multiply(w2, h2)
    area_intersect = np.multiply(w_intersect, h_intersect)

    area_union = area1 + area2 - area_intersect

    return area_union != 0 and area_intersect / area_union > iou_threshold


@numba.njit
def euclidean_distance(box1, box2, distance_threshold):
    """
    Filter based on the euclidean distance between both boxes.
    :param box1: First box for comparison (x, y, width, height).
    :param box2: Second box for comparison (x, y, width, height).
    :param distance_threshold: Threshold of the distance between both boxes.
    :return: Boolean if boxes are closer than @param distance_threshold.
    """

    dx = box1[0] - box2[0]
    dy = box1[1] - box2[1]

    return np.sqrt(dx * dx + dy * dy) < distance_threshold


@numba.jit(nopython=True)
def get_box_vertices(box):
    """Return the vertices of a box.
    :param box: 2D Box with coordinates (x, y, width, height, rotation).
    :return: Vertices of the provided box (absolute coordinates).
    """
    x = box[0]
    y = box[1]
    w = box[2]
    h = box[3]
    yaw = box[4]

    # box vertices without rotation, centered in [0, 0]
    vertices = np.zeros((4, 2), dtype=np.float32)
    vertices[0, 0] = -0.5 * w
    vertices[0, 1] = -0.5 * h
    vertices[1, 0] = 0.5 * w
    vertices[1, 1] = -0.5 * h
    vertices[2, 0] = 0.5 * w
    vertices[2, 1] = 0.5 * h
    vertices[3, 0] = -0.5 * w
    vertices[3, 1] = 0.5 * h

    # rotate and translate box vertices
    c = np.cos(yaw)
    s = np.sin(yaw)
    for i in range(4):
        pi_x = vertices[i, 0]
        pi_y = vertices[i, 1]
        vertices[i, 0] = pi_x * c - pi_y * s + x
        vertices[i, 1] = pi_x * s + pi_y * c + y

    return vertices


def get_box_separating_axis(box_vertices):
    """Return the separating axes based on the provided box vertices.
    :param box_vertices: Vertices of the box for which the seperating axes should be returned.
    :return: Two separating axes of the box.
    """

    # a rectangle has two separating axes
    axes = np.zeros((2, 2))
    ax1 = box_vertices[1] - box_vertices[0]
    norm = np.linalg.norm(ax1)
    if norm != 0:
        axes[0] = ax1 / norm

    ax2 = box_vertices[2] - box_vertices[1]
    norm = np.linalg.norm(ax2)
    if norm != 0:
        axes[1] = ax2 / norm

    return axes


@numba.jit(numba.boolean(numba.float32[::1], numba.float32[::1], numba.float32), nopython=True)
def bounding_rectangle_collision(box1, box2, margin_factor):
    """
    Filter based on the collision of two boxes
    :param box1: First box for comparison (x, y, width, height, rotation).
    :param box2: Second box for comparison (x, y, width, height, rotation).
    :param margin_factor: The collision is computed between both boxes reduced/enlarged by the margin (relative).
    :return: Boolean if boxes collide.
    """

    # b1 and b2 are the original boxes reduced/increased by the margin
    b1 = np.array(box1)
    b2 = np.array(box2)
    b1[2] *= max(margin_factor, 0.01)
    b1[3] *= max(margin_factor, 0.01)
    b1[2] *= max(margin_factor, 0.01)
    b1[3] *= max(margin_factor, 0.01)

    vertices_box1 = get_box_vertices(b1)
    vertices_box2 = get_box_vertices(b2)

    separating_axes = np.vstack([get_box_separating_axis(vertices_box1), get_box_separating_axis(vertices_box2)])

    for i in range(len(separating_axes)):
        axis = separating_axes[i]
        min_box1 = np.infty
        max_box1 = -np.infty
        for vertex in vertices_box1:
            v = axis[0] * vertex[0] + axis[1] * vertex[1]
            min_box1 = min(v, min_box1)
            max_box1 = max(v, max_box1)

        min_box2 = np.infty
        max_box2 = -np.infty
        for vertex in vertices_box2:
            v = axis[0] * vertex[0] + axis[1] * vertex[1]
            min_box2 = min(v, min_box2)
            max_box2 = max(v, max_box2)

        separating = not (
                (min_box2 < max_box1 and max_box2 > min_box1) or (min_box1 < max_box2 and max_box1 > min_box2))

        if separating:
            return False

    return True
