from abc import ABCMeta
from typing import Tuple, Optional, List, Dict, Iterable
from functools import partial

import cv2
import sys
import numpy as np
import logging

from inds.inference.models.base_model import BaseModel, InferenceMode
from inds.core.item import Item

logger = logging.getLogger(__name__)

try:
    from albumentations import ReplayCompose
    from inds.processing.augmentations.transform import (OneOfRetainProb, RelativePadding, RelativeCrop,
                                                         RelativeCenterCrop, )

except Exception as e:
    # Custom transformations could not be included (potentially because optional pacakges are not installed)
    # This is not a problem if there are no preprocessing augmentations to be performed.
    logger.info(f'The preprocessing module could not be imported ({e}). Thus, it is impossible to apply the same '
                'preprocessing steps as performed during training. If this feature is requested later on, '
                'IN:DS will report an ImportError.')
    ...


class ImageBaseModel(BaseModel, metaclass=ABCMeta):

    def __init__(self, input_shape: Tuple[int, int, int, int] = (0, 0, 0, 0),
                 inference_mode: InferenceMode = InferenceMode.WHOLE_IMAGE,
                 selected_rois: Optional[List[str]] = None, preprocessing: Optional[Dict] = None):
        self.input_shape: Tuple[int, int, int, int] = input_shape
        self.inference_mode: InferenceMode = inference_mode
        self.selected_rois: List[str] = [] if selected_rois is None else selected_rois
        self.current_roi_order: List[str] = []
        self.preprocessing: Optional[Dict] = preprocessing

        if self.preprocessing and sys.modules.get('inds.processing') is None:
            raise ImportError('This model is configured to apply preprocessing augmentations.'
                              'However, the required preprocessing package could not be loaded.')

    def preprocess_item(self, item: Item) -> Item:
        """
        Preprocess entire item independ

        :param item:
        :return:
        """
        image = item.get('image')
        image_structured = item.get('image_structured')

        if image is None and (image_structured is None or not self.preprocessing):
            return item

        if self.preprocessing:
            try:
                item['image'] = ReplayCompose.replay(self.preprocessing, image=image_structured, keypoints=[], instances=[],
                                                     positions=[])['image']
            except KeyError as e:
                logger.warning(f'Could not replay training augmentations: {e}. '
                               f'The version of the preprocessing library that was used to augment the '
                               f'image might posess an augmentation ({e}), that was not implemented '
                               f'in the currently installed version of the preprocessing software. '
                               f'Please update the preprocessing software if that is the case.')
                raise e from None

        return item

    def preprocess(self, item: Item) -> Optional[np.ndarray]:
        image: Optional[np.ndarray] = item.get('image')
        if image is None:
            return None

        height, width, channels = image.shape

        image_patches: List[np.ndarray] = []

        if self.inference_mode == InferenceMode.WHOLE_IMAGE:
            image_patches.append(image)

        elif self.inference_mode == InferenceMode.ALL_ROIS or self.inference_mode == InferenceMode.SELECTED_ROIS:
            item_rois: Dict = item.get('rois', {})
            selected_rois: Iterable[str] = item_rois.keys() if self.inference_mode == InferenceMode.ALL_ROIS \
                else self.selected_rois
            self.current_roi_order = []

            for roi_name in selected_rois:
                roi = item_rois.get(roi_name)
                if roi is None:
                    continue
                x_left, x_right, y_top, y_bottom = roi
                if 0 <= x_left < x_right <= width and 0 <= y_top < y_bottom <= height:
                    image_patch = image[y_top:y_bottom, x_left:x_right, :]
                    image_patches.append(image_patch)
                    self.current_roi_order.append(roi_name)

        if len(image_patches) == 0:
            return None

        images = np.empty((len(image_patches), self.input_shape[1], self.input_shape[2], self.input_shape[3]),
                          dtype=np.float32)
        for i, image_patch in enumerate(image_patches):
            # resize
            img = cv2.resize(image_patch, (self.input_shape[3], self.input_shape[2]))

            # transpose from HWC to CHW
            img = img.transpose((2, 0, 1))

            # convert to float
            img = img.astype(np.float32, copy=False)

            # normalize: image = (2.0 / 255.0) * image - 1
            images[i] = (2.0 / 255.0) * img - 1

        return images
