from abc import abstractmethod
from typing import Dict
import numpy as np

import inds.inference.matching as matching
from inds.core.config import ConfigFieldTypes
from inds.core.config.configurable import Configurable
from inds.core.decorators import numba


# XXX: if we use numba, use it at compile time.
@numba.jit(nopython=True)
def multi_class_non_maximum_suppression(confidences, boxes, labels, n_classes, max_selections=200,
                                        matching_threshold=0.5, matching_fn=matching.iou):
    """Apply non maximum suppression to the objects of each class separately

    :param confidences: Matrix (numpy) of shape [N, 1] containing the confidence for each box
    :param boxes: Matrix (numpy) of shape [N, 4] or [N, 5] if orientation is included
    :param labels: Matrix (numpy) of shape [N, 1] containing the class for each box
    :param max_selections: Maximum number of selected boxes per class
    :param matching_threshold: Threshold when to merge boxes
    :return: Selected box indices
    """

    all_selected_indices = []

    for i in range(n_classes):
        class_i_indices = np.where(labels == i)[0]
        class_boxes = boxes[class_i_indices]
        class_confidences = confidences[class_i_indices]

        if class_i_indices.shape[0] > 0:
            selected_indices = non_maximum_suppression(class_confidences, class_boxes, max_selections,
                                                       matching_threshold, matching_fn)
            all_selected_indices.extend(class_i_indices[selected_indices])

    return all_selected_indices


# @numba.jit(numba.int32[::1](numba.float32[::1], numba.float32[:,::1], int32, float32, ?), nopython=True, parallel=True)
def non_maximum_suppression(confidences, boxes, max_selections=200, matching_threshold=0.5, matching_fn=matching.iou):
    """Apply non maximum suppression to aggregate highly overlapping boxes.
    :param confidences: Matrix (numpy) of shape [N, 1] containing the confidence for each box
    :param boxes: Matrix (numpy) of shape [N, 4]
    :param max_selections: Maximum number of selected boxes
    :param matching_threshold: Threshold when to merge boxes
    :param matching_fn: Function to use for comparing two boxes. Use IoU as default.
    :return: Selected box indices
    """

    # sort the confidences(descending)
    sorted_indices = np.argsort(confidences)[::-1]
    n_select = min(boxes.shape[0], max_selections)
    selected_indices = np.empty(n_select, dtype=np.int32)

    n_selected: int = 0
    for i in sorted_indices:
        select = True

        # iterate backwards through selected boxes (overlapping boxes are likely to have similar scores)
        # see implementation of NMS in tensorflow/core/kernels/non_max_suppression_op.cc
        if n_selected > 0:
            for j in selected_indices[n_selected - 1::-1]:
                if matching_fn(boxes[i], boxes[j], matching_threshold):
                    select = False
                    break

        if select:
            selected_indices[n_selected] = i
            n_selected += 1
            if n_selected >= n_select:
                break

    return selected_indices[:n_selected]


class NmsFilter(Configurable):

    @abstractmethod
    def __call__(self, max_confidences, boxes, labels, n_classes):
        return None

    @classmethod
    def get_config_key(cls):
        return 'nms_filter'

    @classmethod
    def get_config_entries(cls) -> Dict:
        if cls == NmsFilter:
            return dict(name='nms_filter',
                        display_name='Detection merging (non-maximum suppression)',
                        description='Overlapping detections are merged within a postprocessing step. Choose the '
                                    'method and its parameters.',
                        type=ConfigFieldTypes.GROUP_SELECTION,
                        default='nms_no_filter',
                        groups=[f.get_config_entries() for f in NmsFilter.__subclasses__()])
        raise NotImplemented(f'NMS filter {cls.__name__} not fully implemented')


class NmsNoFilter(NmsFilter):

    def __call__(self, max_confidences, boxes, labels, n_classes):
        return range(len(max_confidences))

    @classmethod
    def get_config_key(cls):
        return 'nms_no_filter'

    @classmethod
    def get_config_entries(cls) -> Dict:
        return dict(name=cls.get_config_key(),
                    display_name='Maximum filter only',
                    description='The maximum filter is a very fast method for detection merging and '
                                'always applied before every other method. Choose this option to only '
                                'execute this method.',
                    type=ConfigFieldTypes.GROUP,
                    fields=[])


class NmsIouFilter(NmsFilter):

    def __init__(self, iou_threshold: float):
        super().__init__()
        self.iou_threshold = iou_threshold

    def __call__(self, max_confidences, boxes, labels, n_classes):
        return multi_class_non_maximum_suppression(max_confidences, boxes, labels, n_classes,
                                                   matching_threshold=self.iou_threshold,
                                                   matching_fn=matching.iou)

    @classmethod
    def get_config_key(cls):
        return 'nms_iou_filter'

    @classmethod
    def get_config_entries(cls) -> Dict:
        return dict(name=cls.get_config_key(),
                    display_name='IoU filter',
                    description='Detections are merged based their overlap (intersection over union).',
                    type=ConfigFieldTypes.GROUP,
                    fields=[dict(name='iou_threshold',
                                 display_name='IoU threshold',
                                 type=ConfigFieldTypes.DOUBLE_INPUT,
                                 default=0.5,
                                 description='Overlapping detections are merged if their intersection '
                                             'over union (IoU) exceeds this threshold. The IoU '
                                             'computation does not account for rotated boxes, use '
                                             '"Collision filter" instead.')])


class NmsCollisionFilter(NmsFilter):

    def __init__(self, margin_factor: float):
        super().__init__()
        self.margin_factor: float = margin_factor

    def __call__(self, max_confidences, boxes, labels, n_classes):
        return multi_class_non_maximum_suppression(max_confidences, boxes, labels, n_classes,
                                                   matching_threshold=self.margin_factor,
                                                   matching_fn=matching.bounding_rectangle_collision)

    @classmethod
    def get_config_key(cls):
        return 'nms_collision_filter'

    @classmethod
    def get_config_entries(cls) -> Dict:
        return dict(name=cls.get_config_key(),
                    display_name='Collision filter',
                    description='Detections are merged if the collide. This filter accounts for box '
                                'orientations as well.',
                    type=ConfigFieldTypes.GROUP,
                    fields=[dict(name='margin_factor',
                                 display_name='Margin',
                                 type=ConfigFieldTypes.DOUBLE_INPUT,
                                 default=1.0,
                                 description='The collision box size can be increased (> 1.0) or '
                                             'reduced (< 1.0) by the margin parameter.')])
