# -*- coding: utf-8 -*-
"""
This module handles the incoming data and extracts information from it.

Author: GBC09 / BU05 / SW
SICK AG, Waldkirch
email: techsupport0905@sick.de

Copyright note: Redistribution and use in source, with or without modification, are permitted.

Liability clause: THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import struct
import time

from .BinaryParser import BinaryParser
from .ParserHelper import CameraParameters
from .XMLParser import XMLParser
from ..UnitConversion import convertDistanceToMM
from ..Control import logger


class Data:
    """ Gathers methods to handle the raw data. """

    def __init__(self, xmlParser=None, changedCounter=-1, depthmap=None, polarData=None, checksum='E'):
        self.xmlParser = xmlParser
        self.changedCounter = changedCounter
        self.depthmap = depthmap
        self.polarData2D = polarData
        self.checksum = checksum
        self.corrupted = False

        self.parsing_time_s = 0

    def read(self, dataBuffer, convertToMM = True):
        """
        Extracts necessary data segments and triggers parsing of segments. 
        
        dataBuffer:  The raw data from the device. Obtained for example via getFrame() (Stream.py)
        convertToMM: If this is True, depthmap data is converted to millimeters after reading (can result in floating point values).
                     If this is False, the raw (integer) depth data from the device is used:
                       - Tenth millimeters for Visionary S
                       - Quarter millimeters for Visionary T Mini
                       - Millimeters for Visionary T
        """

        parsing_start_time_s = time.time()

        # first 11 bytes contain some internal definitions
        # code threw following error:
        #   File "..\common\Data.py", line 50, in read
        #     unpack('>IIHB', tempBuffer)
        # TypeError: a bytes-like object is required, not 'str'

        # tempBuffer = dataBuffer[0:11]
        tempBuffer = dataBuffer[0:11]
        (magicword, pkglength, protocolVersion, packetType) = \
            struct.unpack('>IIHB', tempBuffer)
        if magicword != 0x02020202:
            raise ValueError('Corrupted package...')

        logger.debug("Package length: %s", pkglength)
        logger.debug("Protocol version: %s", protocolVersion)  # expected to be == 1
        logger.debug("Packet type: %s", packetType)  # expected to be  == 98

        # next four bytes an id (should equal 1) and
        # the number of segments (should be 3)
        tempBuffer = dataBuffer[11:15]
        (segid, numSegments) = struct.unpack('>HH', tempBuffer)
        logger.debug("Blob ID: %s", segid)  # expected to be == 1
        logger.debug("Number of segments: %s", numSegments)  # expected to be  == 3

        # offset and changedCounter, 4 bytes each per segment
        offset = [None] * numSegments
        changedCounter = [None] * numSegments
        tempBuffer = dataBuffer[15:15 + numSegments * 2 * 4]
        for i in range(numSegments):
            index = i * 8
            (offset[i], changedCounter[i]) = \
                struct.unpack('>II', tempBuffer[index:index + 8])
            offset[i] += 11
        logger.debug("Offsets: %s", offset)  # offset in bytes for each segment
        logger.debug("Changed counter: %s", changedCounter)  # counter for changes in the data

        # first segment describes the data format in XML
        xmlSegment = dataBuffer[offset[0]:offset[1]]
        logger.debug("The whole XML segment:")
        logger.debug(xmlSegment)
        # second segment contains the binary data
        binarySegment = dataBuffer[offset[1]:offset[2]]

        if (numSegments == 3):
            overlaySegment = dataBuffer[offset[2]:pkglength+4+4] # numBytes(magicword) = 4, numBytes(pkglength) = 4
            logger.debug("The whole overlay XML segment:")
            logger.debug(overlaySegment)

        checksum = chr(dataBuffer[pkglength+8])
        if checksum != self.checksum:
          logger.error("Checksum is wrong: %s (expected %s)" % (checksum, self.checksum)) # checksum of whole data
          self.corrupted = True
        else:
          logger.debug("Checksum: %s", checksum) # checksum of whole data
          self.corrupted = False

        # parsing the XML in order to extract necessary image information
        # only parse if something has changed
        if (self.changedCounter < changedCounter[0]):
            logger.debug("XML did change, parsing started.")
            myXMLParser = XMLParser()
            myXMLParser.parse(xmlSegment)
            self.xmlParser = myXMLParser
            self.changedCounter = changedCounter[0]
        else:
            logger.debug("XML did not change, not parsing again.")
            myXMLParser = self.xmlParser

        myBinaryParser = BinaryParser()

        self.hasDepthMap = False
        self.hasPolar2D = False
        self.hasCartesian = False

        if myXMLParser.hasDepthMap:
            logger.debug("Data contains depth map, reading camera params")
            self.hasDepthMap = True
            self.cameraParams = \
                CameraParameters(width=myXMLParser.imageWidth,
                                 height=myXMLParser.imageHeight,
                                 cam2worldMatrix=myXMLParser.cam2worldMatrix,
                                 fx=myXMLParser.fx, fy=myXMLParser.fy,
                                 cx=myXMLParser.cx, cy=myXMLParser.cy,
                                 k1=myXMLParser.k1, k2=myXMLParser.k2,
                                 f2rc=myXMLParser.f2rc)

            # extracting data from the binary segment (distance, intensity
            # and confidence).
            if myXMLParser.stereo:
                numBytesDistance = myXMLParser.imageHeight * \
                                   myXMLParser.imageWidth * \
                                   myXMLParser.numBytesPerZValue
            else:
                numBytesDistance = myXMLParser.imageHeight * \
                                   myXMLParser.imageWidth * \
                                   myXMLParser.numBytesPerDistanceValue
            numBytesIntensity = myXMLParser.imageHeight * \
                                myXMLParser.imageWidth * \
                                myXMLParser.numBytesPerIntensityValue
            numBytesConfidence = myXMLParser.imageHeight * \
                                 myXMLParser.imageWidth * \
                                 myXMLParser.numBytesPerConfidenceValue

            try:
                numBytesFrameNumber = myXMLParser.numBytesFrameNumber
                numBytesQuality = myXMLParser.numBytesQuality
                numBytesStatus = myXMLParser.numBytesStatus
            except AttributeError:
                numBytesFrameNumber = 0
                numBytesQuality = 0
                numBytesStatus = 0

            logger.info("Reading binary segment...")
            myBinaryParser.getDepthMap(binarySegment,
                                       numBytesFrameNumber,
                                       numBytesQuality,
                                       numBytesStatus,
                                       numBytesDistance,
                                       numBytesIntensity,
                                       myXMLParser.numBytesPerIntensityValue,
                                       numBytesConfidence)
            logger.info("...done.")

            if convertToMM:
                myBinaryParser.depthmap.distance = convertDistanceToMM(myBinaryParser.depthmap.distance, myXMLParser)
            self.depthmap = myBinaryParser.depthmap

        if myXMLParser.hasPolar2DData:
            self.hasPolar2D = True
            if (myXMLParser.hasDepthMap):
                myBinaryParser.getPolar2D(myBinaryParser.remainingBuffer, myXMLParser.numPolarValues)
            else:
                myBinaryParser.getPolar2D(binarySegment, myXMLParser.numPolarValues)
            if hasattr(myBinaryParser, 'polardata'):
                self.polarData2D = myBinaryParser.polardata
            else:
                self.hasPolar2D = False
        elif myXMLParser.hasCartesianData:
            self.hasCartesian = True
            if (myXMLParser.hasDepthMap):
                myBinaryParser.getCartesian(myBinaryParser.remainingBuffer)
            else:
                myBinaryParser.getCartesian(binarySegment)
            if hasattr(myBinaryParser, 'cartesianData'):
                self.cartesianData = myBinaryParser.cartesianData
            else:
                self.hasCartesian = False

        self.parsing_time_s = time.time() - parsing_start_time_s


